
import sys, logging, pickle, datetime, os, signal, time, itertools, select, queue, signal
import json
import random
import multiprocessing as mp
import threading as mt
from importlib import reload
from pathlib import Path
from PySide2.QtCore import QCoreApplication, QObject, QThread, Qt, QRunnable, QThreadPool, QMutex, Signal


try:
    from element_types import Record, ProcCMD, GuiCMD, GuiException, PythonicError
except ImportError:    
    from Pythonic.element_types import Record, ProcCMD, GuiCMD, GuiException, PythonicError



class ProcessHandler(QRunnable):

    def __init__(self, element, inputdata, identifier, operator):
        super(ProcessHandler, self).__init__()

        self.element    = element
        self.inputData  = inputdata
        self.identifier = identifier
        self.instance   = None
        self.pid        = None
        self.queue      = None 
        self.operator   = operator
        self.element['Config']['Identifier'] = self.identifier

        logging.debug('ProcessHandler::__init__() called') 


    def run(self):
        logging.debug('ProcessHandler::run() -id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))


        bMP = self.element['Config']['GeneralConfig']['MP']

        if bMP:
            self.return_queue = mp.Queue()
            self.cmd_queue    = mp.Queue()
        else:
            self.return_queue = queue.Queue()
            self.cmd_queue    = queue.Queue()

        try:
            # This affects only first invocation
            module = __import__(self.element['Filename'])

            #logging.warning("Load module first time")

            # Reload to execute possible changes
            module = reload(module) 

            elementCls = getattr(module, 'Element')
            logging.debug('ProcessHandler::run() X5 -id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))

        except Exception as e:
            logging.warning('ProcessHandler::run() - Error loading file - id: 0x{:08x}, ident: {:04d} - {} Error: {}'.format(
                self.element['Id'], self.identifier, self.element['Filename'], e))
            return

        self.instance = elementCls(self.element['Id'], self.element['Config'], self.inputData, self.return_queue, self.cmd_queue)
        result = None

        

        if bMP: ## attach Debugger if flag is set
            self.p_0 = mp.Process(target=self.instance.execute_ex)
            self.p_0.start()
            self.pid = self.p_0.pid
        else:
            self.t_0 = mt.Thread(target=self.instance.execute_ex)
            self.t_0.start()



        result = Record(None, None)
        
        ##################################################################
        #                                                                # 
        #                          MULTITHREADING                        #
        #                                                                #
        ##################################################################

        # Check if it is an intemediate result (result.bComplete)
        # or if the execution was stopped by the user (self.element.bStop)



        while not bMP:

            try:
                # First: Check if there is somethin in the Queue
                result = self.return_queue.get(block=True, timeout=0.2)
                # Seconds: Forward the result (is present)
                self.operator.operationDone(self.element['Id'], self.element['AreaNo'], result, self.identifier)
            except queue.Empty:
                #logging.debug('return_queue empty')
                pass
            # Thirs: Check if Thread is still alive
            if not self.t_0.is_alive():
                break

            #logging.debug('ProcessHandler::run() - Multithreading: result received - id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))



        ##################################################################
        #                                                                # 
        #                         MULTIPROCESSING                        #
        #                                                                #
        ##################################################################

        while bMP:

            try:
                result = self.return_queue.get(block=True, timeout=0.2)

                self.operator.operationDone(self.element['Id'], self.element['AreaNo'], result, self.identifier)
                #logging.debug('ProcessHandler::run() - Multiprocessing: execution completed - id: 0x{:08x}, ident: {:04d}, pid: {}'.format(
                #    self.element['Id'], self.identifier, self.p_0.pid))
            except queue.Empty:
                #logging.debug('return_queue empty')
                pass
            
            if not self.p_0.is_alive():
                break
            
        self.operator.removeOperatorThread(self.element['Id'], self.identifier)
        
    def stop(self):
        logging.debug('ProcessHandler::stop() - id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))
        self.cmd_queue.put(ProcCMD(None, True))

    def feed(self, data): 
        logging.debug('ProcessHandler::feed() - id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))
        """
        if isinstance(data, ProcCMD):
            self.cmd_queue.put(data)
        else:
        """
        self.cmd_queue.put(ProcCMD(data))

class OperatorStartAll(QRunnable):
    

    def __init__(self, operator):
        super(OperatorStartAll, self).__init__()
        self.operator       = operator
        self.setAutoDelete(False)
        
    def run(self):

        logging.debug('OperatorStartAll::startAll() called')
            
        startElements = [x for x in self.operator.currentConfig if not x['Socket']]

        for startElement in startElements:
            
            
            processes = filter(lambda item: item[1].element['Id'] == startElement['Id'], self.operator.processHandles.items())
            runningProcess = next(processes, None)

            if(runningProcess):
                logging.debug('Operator::startAll() -Element already running - {} - id: 0x{:08x}'.format(
                    runningProcess[1].element['ObjectName'], runningProcess[1].element['Id']))
                return
            else:
                logging.debug('Operator::startAll() -Element started - {} - id: 0x{:08x}'.format(
                   startElement['ObjectName'], startElement['Id']))
                
                self.operator.createProcHandle(startElement)

class OperatorElementOpDone(QRunnable):

    def __init__(self, config, id, record : Record, identifier, operator):
        super(OperatorElementOpDone, self).__init__()


        self.currentConfig  = config
        self.id             = id
        self.record         = record
        self.identifier     = identifier
        self.operator       = operator


    def run(self):

        logging.debug("OperatorElementOpDone::run() called")

        cfgElement = [x for x in self.currentConfig if x['Id'] == self.id][0]



        if cfgElement['Config']['GeneralConfig']['Logging'] and self.record.message: # Log Message enabled
            if isinstance(self.record.data, PythonicError):
                logging.warning('{} - {}'.format(cfgElement['ObjectName'], self.record.message))
            else:
                logging.info('{} - {}'.format(cfgElement['ObjectName'], self.record.message))

            data = {
                'Id'        : cfgElement['Id'],
                'ObjectName': cfgElement['ObjectName'],
                'Message'   : self.record.message
            }
            address = {
                'target'    : 'MainWindow',                        
            }
            cmd = { 
                'cmd'       : 'ElementMessage',
                'address'   : address,
                'data'      : data
            }

            self.operator.emitCommand(cmd)


        if cfgElement['Config']['GeneralConfig']['Debug'] : # Log Output enabled

            data = {
                'Id'        : cfgElement['Id'],
                'AreaNo'    : cfgElement['AreaNo'],
                'ObjectName': cfgElement['ObjectName'],
                'Output'    : str(self.record.data) 
            }
            address = {
                'target'    : 'MainWindow',                        
            }
            cmd = { 
                'cmd'       : 'DebugOutput',
                'address'   : address,
                'data'      : data
            }

            self.operator.emitCommand(cmd)

        # return if the element has no childs
        if not cfgElement['Childs']:
            return
        
        for childId in cfgElement['Childs']:
            childElement = [x for x in self.currentConfig if x['Id'] == childId][0]
            self.operator.createProcHandle(childElement, self.record.data)
            self.operator.highlightConnection(self.id, childId, childElement['AreaNo'])

class OperatorCreateProcHandle(QRunnable):
    
    def __init__(self, element, inputData, procHandles, operator):
        super(OperatorCreateProcHandle, self).__init__()
        self.element        = element
        self.inputData      = inputData
        self.procHandles    = procHandles
        self.operator       = operator

    def run(self):

        # check if inputData is a Stop command
        bStop = False
        if isinstance(self.inputData, ProcCMD):
            bStop = self.inputData.bStop



        # Check if an already running element has to be stopped
        if self.element['AllowStream'] or bStop:
            
            runningInstance =  list(filter(lambda item: item[1].element['Id'] == self.element['Id'], self.procHandles.items())) 
            # when runningInstance contains an element forward input data to it
            if runningInstance:
                # [0] first element in the list, [1] ProcessHandler
                tagetProc = runningInstance[0][1]

                if not bStop:
                    tagetProc.feed(self.inputData)
                else:
                    tagetProc.stop() # Send Stop command to running process

                return


        # return here if inputData is a Stop command
        if bStop:
            return

        identifier = self.operator.getIdent()
        runElement = ProcessHandler(self.element, self.inputData, identifier, self.operator)

        if self.element["HighlightState"]: 
            self.operator.updateStatus(self.element, True)
        
        self.operator.addHandle(identifier, runElement)
        self.operator.threadpool.start(runElement)
        
        logging.debug('Operator::createProcHandle() called - identifier: {:04d}'.format(identifier))

class OperatorReturnElementState(QRunnable):

    def __init__(self, processHandles, operator):
        super(OperatorReturnElementState, self).__init__()
        self.processHandles = processHandles
        self.operator       = operator

    def run(self):

        for threadIdentifier, processHandle in self.processHandles.items():
            #logging.debug('running elements: 0x{:08x}'.format(processHandle.element['Id']))
            self.operator.updateStatus(processHandle.element, True)

class OperatorStopExec(QRunnable):

    def __init__(self, processHandles, id, operator):
        super(OperatorStopExec, self).__init__()
        self.processHandles = processHandles
        self.operator       = operator
        self.id             = id

    def run(self):

        for threadIdentifier, processHandle in self.processHandles.items():
            if processHandle.element['Id'] == self.id:
                processHandle.stop()

class Operator(QObject):

    currentConfig   = None
    processHandles  = {}
    command         = Signal(object) # command

    def __init__(self):
        super(Operator, self).__init__()


        logging.debug('Operator::__init__() called')
        self.threadpool         = QThreadPool.globalInstance()
        self.procHandleMutex    = QMutex()

        self._startAll          = OperatorStartAll(self)

        self.identGenMutex      = QMutex()
        self.n_ident            = 0


    def start(self, config):

        logging.info('<#>DAEMON STARTED<#>')

        if not config: # return here when there is no config file
            return

        self.currentConfig = config

        # check for autostart elements

        startElements = [x for x in self.currentConfig if not x['Socket'] and x['Config']['GeneralConfig']['Autostart']]

        for startElement in startElements:
            logging.info("Autostart " + startElement['ObjectName'])
            self.createProcHandle(startElement)

    def getIdent(self):

        self.identGenMutex.lock()
        
        self.n_ident += 1 
        # check if current ident is already in use
        while list(filter(lambda item: item[0] == self.n_ident, self.processHandles.items())):
            self.n_ident += 1
        

        if not self.n_ident & 0x7fff:
            self.n_ident = 0
        self.identGenMutex.unlock()

        return self.n_ident

    def startExec(self, id, config):
        logging.debug('Operator::startExec() called - id: 0x{:08x}'.format(id))
        ## create processor and forward config and start filename

        self.currentConfig = config
        # https://stackoverflow.com/questions/34609935/passing-a-function-with-two-arguments-to-filter-in-python

        # return first element which matches the ID
        startElement = [x for x in config if x['Id'] == id][0]

        self.createProcHandle(startElement)

    def createProcHandle(self, element, inputData=None):    
        #logging.debug('Operator::startExec() called - id: 0x{:08x}'.format(id))

        self.procHandleMutex.lock()
        processHandles = self.processHandles.copy()
        self.procHandleMutex.unlock()

        procHandle = OperatorCreateProcHandle(element, inputData, processHandles, self)

        self.threadpool.start(procHandle)    

    def addHandle(self, identifier, handle):
        
        self.procHandleMutex.lock()
        self.processHandles[identifier] = handle
        self.procHandleMutex.unlock()

    def stopExec(self, id):
        logging.debug('Operator::stopExec() called - id: 0x{:08x}'.format(id))

        self.procHandleMutex.lock()
        processHandles = self.processHandles.copy()
        self.procHandleMutex.unlock()

        stopOperator = OperatorStopExec(processHandles, id, self)

        self.threadpool.start(stopOperator)

    def stopAll(self):

        logging.debug('Operator::stopAll() called')
        logging.info('User command: Stop All')
        self.procHandleMutex.lock()
        for threadIdentifier, processHandle in self.processHandles.items():
            processHandle.stop()
        self.procHandleMutex.unlock()

    def startAll(self, config):

        logging.info('User command: Start All')
        self.currentConfig = config
        self.threadpool.start(self._startAll)

    def killAll(self):
        
        logging.debug('Operator::killAll() called')
        logging.info('User command: Kill All Processes')   
        # Separate dict must be created because call to removerOperatorThreads
        # modifies self.processHandles
        self.procHandleMutex.lock()
        processes = dict(filter(lambda item: item[1].pid, self.processHandles.items())) 
        self.procHandleMutex.unlock()
        for threadIdentifier, processHandle in processes.items():
            os.kill(processHandle.pid, signal.SIGTERM)
            # removeOperatorThread is called in run() function of ProcessHandler
        
    def getElementStates(self):
        
        logging.debug('Operator::getElementStates() called')
        
        self.procHandleMutex.lock()
        processHandles = self.processHandles.copy()
        self.procHandleMutex.unlock()

        stateOperator = OperatorReturnElementState(processHandles, self)

        self.threadpool.start(stateOperator)
        
    def updateStatus(self, element, status):
        #start highlight
        # area
        # id
        # target = "Element"
        # cmd = UpdateElementStatus
        #logging.debug('Operator::updateStatus() called - {} - id: 0x{:08x}'.format(status, element['Id']))

        address = {
            'target'    : 'Element',
            'area'      : element['AreaNo'],
            'id'        : element['Id']              
        }
        
        cmd = { 
            'cmd'       : 'UpdateElementStatus',
            'address'   : address,
            'data'      : status
            }

        self.command.emit(cmd)

    def highlightConnection(self, parentId, childId, wrkArea):

        logging.debug('Operator::updateStatus() called')
        address = {
            'target'    : 'WorkingArea',
            'area'      : wrkArea           
        }

        data = {
            'parentId'  : parentId,
            'childId'   : childId
        }
        
        cmd = { 
            'cmd'       : 'HighlightConnection',
            'address'   : address,
            'data'      : data
            }

        self.command.emit(cmd)

    def emitCommand(self, command):

        self.command.emit(command)

    def operationDone(self, id, area, record, identifier):

        logging.debug('Operator::operationDone() result received - id: 0x{:08x}, ident: {:04d}'.format(id, identifier))

        if isinstance(record, GuiCMD):

            address = {
                'target'    : 'Element',  
                'id'        : id,
                'area'      : area            
            }
            cmd = { 
                'cmd'       : 'ElementText',
                'address'   : address,
                'data'      : record.text
            }
            self.command.emit(cmd)
            return

        if isinstance(record, GuiException):

            address = {
                'target'    : 'Element',  
                'id'        : id,
                'area'      : area            
            }
            cmd = { 
                'cmd'       : 'ElementException',
                'address'   : address
            }
            # log critical exceptions from multiprocessed elements
            logging.error(repr(record.e))
            self.command.emit(cmd)
            return
        
        operationDoneRunnable = OperatorElementOpDone(self.currentConfig, id, record, identifier, self)

        self.threadpool.start(operationDoneRunnable)
        
    def removeOperatorThread(self, id, identifier):
        
        logging.debug('Operator::removeOperatorThread() called - id: 0x{:08x}, ident: {:04d}'.format(id, identifier))

        self.procHandleMutex.lock()
        procHandle = self.processHandles[identifier]
        if procHandle.element["HighlightState"]:
            self.updateStatus(procHandle.element, False)

        del self.processHandles[identifier]

        self.procHandleMutex.unlock()


import queue
from telegram import Update
from telegram.ext import Updater, CommandHandler, CallbackContext, MessageHandler, Filters

try:
    from element_types import Record, Function, ProcCMD, GuiCMD, SetPersist, PythonicError
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD, SetPersist, PythonicError

    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):

        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################

        cmd = None
        specificConfig = self.config.get('SpecificConfig')
        chat_ids = SetPersist('chat_ids')

        if not specificConfig:

            recordDone = Record(None, message='Config missing')
            self.return_queue.put(recordDone)
            return

        token = next(attr['Data'] for attr in specificConfig if attr['Name'] == 'Token')

        if not token:
            recordDone = Record(None, message='Token missing')
            self.return_queue.put(recordDone)
            return

        updater = Updater(token=self.config['SpecificConfig'][0]['Data'], use_context=True)

        dispatcher = updater.dispatcher

        def start(update: Update, context: CallbackContext):
            
            chat_ids.add(update.message.chat_id) 
            context.bot.send_message(chat_id=update.effective_chat.id, text="Hello, this chat ID is now registered for communication.")

        def unknown(update, context):
            context.bot.send_message(chat_id=update.effective_chat.id, text='Sorry, I didn\'t understand that command.')

        def message(update, context):
            context.bot.send_message(chat_id=update.effective_chat.id, text='Message received')
            guitext = GuiCMD('Message received from: {}, MsgId.: {:d}'.format(update.message.from_user.first_name, update.message.message_id))
            self.return_queue.put(guitext)
            record = Record(update.message, 'Message received from: {}, MsgId.: {:d}'.format(update.message.from_user.first_name, update.message.message_id))
            self.return_queue.put(record)

        start_handler       = CommandHandler('start', start)
        message_handler     = MessageHandler(Filters.text &~ Filters.command, message)
        unknown_cmd_handler = MessageHandler(Filters.command, unknown)

        dispatcher.add_handler(start_handler)
        dispatcher.add_handler(message_handler) # muss als letztes hinzugefügt werden
        dispatcher.add_handler(unknown_cmd_handler)

        updater.start_polling()

        while(updater.running):

            try:
                # Block for 1 second and wait for incoming commands 
                cmd = self.cmd_queue.get(block=True, timeout=1)
            except queue.Empty:
                pass

            if isinstance(cmd, ProcCMD) and cmd.bStop:
                # Stop command received, exit
                updater.stop()

            elif isinstance(cmd, ProcCMD):

                guitext = GuiCMD("Sending data: " + str(cmd.data))
                self.return_queue.put(guitext)

                for chat_id in chat_ids.copy():
                    try:
                        dispatcher.bot.send_message(chat_id=chat_id, text=str(cmd.data))
                    except Exception as e:
                        chat_ids.discard(chat_id)
                        self.return_queue.put(Record(PythonicError(e), 'Error sending message, related chat id removed'))

            cmd = None


import time, queue
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
from gpiozero import Button

class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):

        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################
        
        specificConfig = self.config.get('SpecificConfig')

        if not specificConfig:
            self.return_queue.put(Record(None, message='Trigger: {:04d}'.format(self.config['Identifier'])))
            return

        gpioName        = None
        mainMode        = None
        cmd             = None

        self.gpio       = None
        self.initFlag   = False

        for attrs in specificConfig:
            if attrs['Name'] == 'GPIO':
                gpioName = attrs['Data']
            if attrs['Name'] == 'MainMode':
                mainMode = attrs['Data']


        self.gpio = Button(gpioName)

        if mainMode == 'Fire when released':

            self.gpio.when_released = self.onReleased

        elif mainMode == 'Fire when pressed': 

            self.gpio.when_pressed = self.onPressed



        #####################################
        #                                   #
        #     Start of the infinite loop    #
        #                                   #
        #####################################

        while(True):

            # Example code: Do something

            try:
                # Block for 1 second and wait for incoming commands 
                cmd = None
                cmd = self.cmd_queue.get(block=True, timeout=1)
            except queue.Empty:
                pass

            if isinstance(cmd, ProcCMD):
                if cmd.bStop:
                    # Stop command received, exit 
                    self.return_queue.put(Record(None, 'GPIO{} closed'.format(self.gpio.pin.number))) 
                    self.gpio.close()
                    return


    def onPressed(self):

        self.return_queue.put(Record(None, 'GPIO{} pressed'.format(self.gpio.pin.number))) 

    def onReleased(self):

        self.return_queue.put(Record(None, 'GPIO{} released'.format(self.gpio.pin.number))) 


import time, queue
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):


        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################


        cmd = None
        cnt = 0

        #####################################
        #                                   #
        #     Start of the infinite loop    #
        #                                   #
        #####################################

        # The example executes an infinite loop till it's receives a stop command
        while(True):

            # Example code: Do something

            try:
                # Block for 1 second and wait for incoming commands 
                cmd = self.cmd_queue.get(block=True, timeout=1)
            except queue.Empty:
                pass

            if isinstance(cmd, ProcCMD):
                if cmd.bStop:
                    # Stop command received, exit
                    return
                else:
                    # Example Code: Send number of received data packets to GUI
                    cnt+=1
                    guitext = GuiCMD('Data received: {}'.format(cnt))
                    self.return_queue.put(guitext)
                    cmd.data += 1
                    
                    # Example Code: Increment received data by one and forward it to subsequent elements
                    recordDone = Record(cmd.data, 'Sending value of cnt: {}'.format(cmd.data))     
                    self.return_queue.put(recordDone)
                    
                    cmd = None























class ProcessHandler(QThread):

    execComplete  = Signal(object, object, object) # id, data, thread-identifier
    removeSelf    = Signal(object, object) # id, thread-identifier

    def __init__(self, element, inputdata, identifier, parent=None):
        #super().__init__()
        super(ProcessHandler, self).__init__(parent)
        self.element    = element
        self.inputData  = inputdata
        self.identifier = identifier
        self.instance   = None
        self.element['Config']['Identifier'] = self.identifier
        self.pid        = None
        self.queue      = None 


        self.finished.connect(self.done)

    def run(self):
        #logging.debug('ProcessHandler::run() -id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))
        bMP = self.element['Config']['GeneralConfig']['MP']
        

        if bMP:
            self.return_queue = mp.Queue()
            self.cmd_queue    = mp.Queue()
        else:
            self.return_queue = queue.Queue()
            self.cmd_queue    = queue.Queue()

        try:

            # This affects only first invocation
            module = __import__(self.element['Filename'])
            #logging.warning("Load module first time")
            
            # Reload to execute possible changes
            module = reload(module) 

            elementCls = getattr(module, 'Element')

        except Exception as e:
            logging.warning('ProcessHandler::run() - Error loading file - id: 0x{:08x}, ident: {:04d} - {} Error: {}'.format(
                self.element['Id'], self.identifier, self.element['Filename'], e))
            return

        self.instance = elementCls(self.element['Config'], self.inputData, self.return_queue, self.cmd_queue)
        result = None

        

        if bMP: ## attach Debugger if flag is set
            self.p_0 = mp.Process(target=self.instance.execute)
            self.p_0.start()
            self.pid = self.p_0.pid
        else:
            self.t_0 = mt.Thread(target=self.instance.execute)
            self.t_0.start()



        result = Record(None, None)
        
        ##################################################################
        #                                                                # 
        #                          MULTITHREADING                        #
        #                                                                #
        ##################################################################

        # Check if it is an intemediate result (result.bComplete)
        # or if the execution was stopped by the user (self.element.bStop)
        
        while not bMP:

            try:
                # First: Check if there is somethin in the Queue
                result = self.return_queue.get(block=True, timeout=0.2)
                # Seconds: Forward the result (is present)
                self.execComplete.emit(self.element['Id'], result, self.identifier)
            except queue.Empty:
                #logging.debug('return_queue empty')
                pass
            # Thirs: Check if Thread is still alive
            if not self.t_0.is_alive():
                break

            #logging.debug('ProcessHandler::run() - Multithreading: result received - id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))



        ##################################################################
        #                                                                # 
        #                         MULTIPROCESSING                        #
        #                                                                #
        ##################################################################

        while bMP:

            try:
                result = self.return_queue.get(block=True, timeout=0.2)
                self.execComplete.emit(self.element['Id'], result, self.identifier)

                #logging.debug('ProcessHandler::run() - Multiprocessing: execution completed - id: 0x{:08x}, ident: {:04d}, pid: {}'.format(
                #    self.element['Id'], self.identifier, self.p_0.pid))
            except queue.Empty:
                #logging.debug('return_queue empty')
                pass
            
            if not self.p_0.is_alive():
                break
            


        #logging.debug('ProcessHandler::run() - PROCESSING DONE - id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))
        
    def stop(self):
        logging.debug('ProcessHandler::stop() - id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))
        self.cmd_queue.put(ProcCMD(True))

    def done(self):
        #logging.debug('ProcessHandler::done() removing Self - id: 0x{:08x}, ident: {:04d}'.format(self.element['Id'], self.identifier))
        self.removeSelf.emit(self.element['Id'], self.identifier)


import queue
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):


        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################

        recordDone = Record(ProcCMD(None, bStop=True), 'Sending Stop command')     
        self.return_queue.put(recordDone)


import time, queue
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
from gpiozero import LED, PWMLED

class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):

        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################
        specificConfig = self.config.get('SpecificConfig')

        if not specificConfig:
            self.return_queue.put(Record(None, message='Trigger: {:04d}'.format(self.config['Identifier'])))
            return

        gpioName        = None
        mainMode        = None
        subModeLED      = None
        subModePWMLED   = None
        gpioWorker      = None
        cmd             = None

        self.gpio       = None
        self.initFlag   = False

        for attrs in specificConfig:
            if attrs['Name'] == 'GPIO':
                gpioName = attrs['Data']
            if attrs['Name'] == 'MainMode':
                mainMode = attrs['Data']
            elif attrs['Name'] == 'SubModeLED':
                subModeLED = attrs['Data']
            elif attrs['Name'] == 'SubModePWMLED':
                subModePWMLED = attrs['Data']




        if mainMode == 'LED':

            self.gpio = LED(gpioName, initial_value=False)

            if subModeLED == 'Toggle on input':

                gpioWorker = self.ledWorkerToggle
                self.gpio.toggle()
                self.logLEDstate()

            elif subModeLED == 'Control on Input':

                gpioWorker = self.ledWorkerCtrl
                # set initial state
                if self.inputData is not None:
                    if self.inputData:
                        self.gpio.on()
                    else:
                        self.gpio.off()

                    self.logLEDstate()

            elif subModeLED == 'Blink':
                def a(cmd = None): pass
                gpioWorker = a # assign an empty function
                self.gpio.blink()
                self.return_queue.put(Record(None, 'Start LED Blink Mode on GPIO{}'.format(self.gpio.pin.number))) 



        elif mainMode == 'PWMLED': 

            self.gpio = PWMLED(gpioName, initial_value=False)

            if subModePWMLED == 'Control on Input':

                gpioWorker = self.pwmLedWorkerCtrl

                if self.inputData is not None:
                    self.gpio.value = self.inputData
                    self.return_queue.put(Record(None, 'PWMLED: Set brightness on GPIO{} to {:.2f}'.format(self.gpio.pin.number, self.inputData))) 

            elif subModePWMLED == 'Pulse':
                def a(cmd = None): pass
                gpioWorker = a # assign an empty function
                self.gpio.pulse()
                self.return_queue.put(Record(None, 'Start PWMLED Pulse Mode on GPIO{}'.format(self.gpio.pin.number))) 

        #####################################
        #                                   #
        #     Start of the infinite loop    #
        #                                   #
        #####################################

        while(True):

            # Example code: Do something

            try:
                # Block for 1 second and wait for incoming commands 
                cmd = None
                cmd = self.cmd_queue.get(block=True, timeout=1)
            except queue.Empty:
                pass

            if isinstance(cmd, ProcCMD):
                if cmd.bStop:
                    # Stop command received, exit 
                    self.return_queue.put(Record(None, 'GPIO{} closed'.format(self.gpio.pin.number))) 
                    self.gpio.close()
                    return

            gpioWorker(cmd)

    def logLEDstate(self):   
        self.return_queue.put(Record(None, 'Switch LED on GPIO{} to {}'.format(self.gpio.pin.number, self.gpio.is_active))) 

    def ledWorkerToggle(self, cmd = None):

        if cmd is None:
            return

        self.gpio.toggle()
        self.logLEDstate()


    def ledWorkerCtrl(self, cmd = None):

        if cmd is None:
            return

        if cmd.data:
            self.gpio.on()
        else:
            self.gpio.off()

        self.logLEDstate()

    def pwmLedWorkerCtrl(self, cmd = None):

        if cmd is None:
            return
        try:
            self.gpio.value = cmd.data
        except Exception:
            self.gpio.close()
            raise
        self.return_queue.put(Record(None, 'PWMLED: Set brightness on GPIO{} to {:.2f}'.format(self.gpio.pin.number, cmd.data)) ) 


import ccxt

sId = 'binance'

exchangeClass = getattr(ccxt, sId)

# https://github.com/ccxt/ccxt/wiki/Manual#overview
# Ein Element mit Schalter für Public/Private
"""
exchange = exchangeClass({
    'apiKey'            : 'YPUR_API_KEY', # text
    'secret'            : 'SECRET_KEY', #text
    'timeout'           : 30000, # milliseconds unsigned int
    'enableRateLimit'   : True,  # checkbox
})
"""
exchange = exchangeClass({'enableRateLimit'   : True})
exchange = ccxt.binance()

#tickers = exchange.fetch_tickers() # keine argumente
#if exchange.has['fetchOHLCV']: #https://github.com/ccxt/ccxt/wiki/Manual#ohlcv-candlestick-charts
#def myfunc2(*args, **kwargs): hier mit arbeiten


# Public API 

lmarkets = exchange.load_markets()
fmarkets = exchange.fetch_markets()
currencies = exchange.fetch_currencies()
ticker = exchange.fetch_ticker('ETH/BTC')
tickers = exchange.fetch_tickers('ETH/BTC XMR/BTC ADA/BNB')
orderbook = exchange.fetch_order_book('ADA/BNB')

ohlcv = exchange.fetch_ohlcv('ADA/BNB', timeframe='1m', limit=500)
exchange.fetch_ohlcv()
#since = UTC timestamp in milliseconds

#status = exchange.fetch_status()

#trades = exchange.fetch_trades('ADA/BNB')


# Private API

balance = = exchange.fetch_balance()

order = exchange.create_order('ETH/BTC', 'Market/limit/stoploss/stoplosslimit/', 'buy/sell', 123.455, price)

fetch_orders = exchange.fetch_orders('ETH/BTC')

fetch_open_orders = exchange.fetch_open_orders('ETH/BTC')

fetch_closed_orders = exchange.fetch_closed_orders('ETH/BTC')

my_trades = exchange.fetch_my_trades('ETH/BTC')

withdraw = exchange.withdraw('BTC', amount, adress)
exchange.withdraw()

print('end')


import logging, datetime
from pathlib import Path
from PySide2.QtCore import QObject, QTimer, Signal



class LogFileHandler(QObject):

    update_logdate = Signal(str) # command

    def __init__(self, log_lvl):
        super(LogFileHandler, self).__init__()
        logging.debug('LogFileHandler::__init__() called')
        self.log_lvl    = log_lvl
        self.log_path   = Path.home() / 'Pythonic' / 'log'
        self.log_date   = datetime.datetime.now()
        self.formatter  = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(message)s', datefmt='%H:%M:%S')
        self.logger     = logging.getLogger()

        # Apply log_lvl
        self.logger.setLevel(self.log_lvl)
        self.logger.propagate = False
        # Remove default Stream Handler
        self.logger.removeHandler(self.logger.handlers[0])

        # Create directory structure for logging
        self.log_date_str = self.log_date.strftime('%Y_%m_%d')

        file_path = '{}/{}.txt'.format(str(self.log_path), self.log_date_str) 

        # Setup logger

        file_handler = logging.FileHandler(file_path)
        file_handler.setLevel(self.log_lvl)
        file_handler.setFormatter(self.formatter)

        self.logger.addHandler(file_handler)


        self.timer = QTimer()
        self.timer.start(1000)
        self.timer.timeout.connect(self.checkLogDate)

    def checkLogDate(self):

        #print("CHECK")
        
        now = datetime.datetime.now().date()
        
        if (now != self.log_date.date()):
            logging.info('CheckTime::run() - Changing logfile')
            self.logger.removeHandler(self.logger.handlers[0])
            self.log_date_str = now.strftime('%Y_%m_%d')
            file_path = '{}/{}.txt'.format(str(self.log_path), self.log_date_str) 
            file_handler = logging.FileHandler(file_path)
            file_handler.setLevel(self.log_lvl)
            file_handler.setFormatter(self.formatter)
            self.logger.addHandler(file_handler)
            self.log_date = datetime.datetime.now()
            self.update_logdate.emit(self.log_date_str)
        


import os, logging, json
from copy import deepcopy
from shutil import copyfile
from datetime import datetime
from pathlib import Path
from PySide2.QtCore import QThread, QObject, Signal, QMutex


class ExecSysCMD(QThread):

    cmd = None

    def __init__(self):
        super().__init__()

    def execCommand(self, cmd):
        self.cmd = cmd
        self.start()

    def run(self):

        logging.debug('ExecSysCMD::run() called')

        if os.name == 'nt' and self.cmd['Win32'] != '':
            os.system(self.cmd['Win32'])
        elif self.cmd['Unix'] != '':
            os.system(self.cmd['Unix'])


class ConfigWriter(QThread):

    config      = None
    configSaved = Signal(object)   
    mutex       = QMutex()

    def __init__(self):
        super().__init__()
        self.cfg_file = Path.home() / 'Pythonic' / 'current_config.json'

    def saveConfig(self, config):

        self.mutex.lock()
        self.config = deepcopy(config)
        self.mutex.unlock()
        self.start()
    
    def run(self):

        logging.debug('ConfigWriter::saveConfig() called')

        self.mutex.lock()
        try:
            with open( self.cfg_file, 'w') as file:
                json.dump(self.config, file, indent=4)
        except Exception as e:
            logging.debug('ConfigWriter::saveConfig() - {}'.format(e))
            pass
            return
            


        self.mutex.unlock()
        last_saved = "Config last saved: " + datetime.now().strftime('%H:%M:%S')

        cmd = {  'cmd'       : 'SetInfoText',
                 'data'      : last_saved,
                 'address'   : {"target" : "MainWindow"}}

        self.configSaved.emit(cmd)


class EditorLoaderThread(QThread):

    editorLoaded        = Signal(object)

    def __init__(self, address, typeName):
        super().__init__()

        self.address    = address
        self.typeName   = typeName + '.editor'
        self.editorPath = Path(__file__).parent.absolute() / 'public_html' / 'config' / 'Toolbox'


    def run(self):

        logging.debug('EditorLoader::run() called')
        config = None
        bFound = False

        for dirpath, dirnames, filenames in os.walk(self.editorPath):
            if self.typeName in filenames:

                try:
                    with open(os.path.join(dirpath ,self.typeName), 'r') as file:

                        config = json.load(file)

                
                    cmd = { 'cmd' : 'ElementEditorConfig',
                            'address'   : self.address,
                            'data'      : config }
                    
                    logging.debug('EditorLoader::run() config loaded')
                    bFound = True
                    self.editorLoaded.emit(cmd)
                    #sleep(1)

                except Exception as e:
                    logging.warning('EditorLoader::run() - error opening file: {}'.format(e))
                
                
                break
            
                

        if not bFound:
            logging.warning('EditorLoader::run() - editor config file {} not found'.format(self.typeName))


class EditorLoader(QObject):

    editorLoaded        = Signal(object)

    typeName    = None
    address     = None

    threadList = []

    def __init__(self):
        super().__init__()
        self.cwd = os.path.dirname(__file__)

    def startLoad(self, address, typeName):

        logging.debug('EditorLoader::startLoad() - called')

        newThread   = EditorLoaderThread(address, typeName)
        newThread.editorLoaded.connect(self.fwrdCmd)
        newThread.finished.connect(self.cleanupThreadList)


        self.threadList.append(newThread)

        newThread.start()

    def fwrdCmd(self, cmd):
        self.editorLoaded.emit(cmd)

    def cleanupThreadList(self):
        # Remove finished threads from memory
        #i = 5

        # BAUSTELLE
        # If the threads are not removed from the list it will cause a memory leak
        # If the threads are removed from list it will cause a QThread Exception:
        # QThread: Destroyed while thread is still running
        #self.threadList[:] = [thread for thread in self.threadList if not thread.isRunning() ]
        for thread in self.threadList:
            x = thread
            if not thread.isRunning():
                self.threadList.remove(thread)
                x.editorLoaded.disconnect(self.fwrdCmd)
                x.finished.disconnect(self.cleanupThreadList)
                x.deleteLater()


class ToolboxLoader(QThread):

    tooldataLoaded      = Signal(object)

    def __init__(self):
        super().__init__()
        self.toolBoxPath = Path(__file__).parent.absolute() / 'public_html' / 'config' / 'Toolbox'

    def run(self):

        logging.debug('ToolboxLoader::run() called')
        
        #toolDirs = [ f for f in os.scandir(os.path.join(self.cwd + self.www_config + '/Toolbox/')) if f.is_dir() ]
        toolDirs = [ f for f in os.scandir(self.toolBoxPath) if f.is_dir() ]

        elements = [(d, f) for d in toolDirs for f in os.listdir(d.path) if f.endswith('.json')]
        elements.sort(key=lambda record: record[1]) # sort elements alphabetical
        elements.sort(key=lambda record: record[0].name) # sort groups alphabetical
        elementsJSON = []
        
        for d, f in elements:
            
            filepath = os.path.join(d.path, f)
            e = None
            with open(filepath, 'r') as file:
                e = json.load(file)

            element = { 'assignment' : d.name[1:],
                        'config' : e}

            elementsJSON.append(element)
            #logging.debug('MainWorker::loadTools() called')

        address = { 'target' : 'MainWindow'}

        cmd = { 'cmd'       : 'Toolbox',
                'address'   : address,
                'data'      : elementsJSON }
        
        self.tooldataLoaded.emit(cmd)


class ConfigLoader(QThread):

    configLoaded      = Signal(object)

    def __init__(self):
        super().__init__()
        self.config     = None
        self.home_path  = Path.home() / 'Pythonic'
        self.cfg_file   = self.home_path  / 'current_config.json'

    def loadConfigSync(self):

        try:
            with open(self.cfg_file, 'r') as file:
                config = json.load(file)

            return config


        except Exception as e:

            logging.warning('No config file found')
            return None


    def run(self):

        self.loadConfig()


    def loadConfig(self, bRecover=False):

        try:
            with open(self.cfg_file, 'r') as file:
                config = json.load(file)

            self.configLoaded.emit(config)

            if(bRecover):
                logging.warning('Old config restored successfully')

        except Exception as e:
            
            # Return if an exception occured when already tried
            # to recover the old config
            if(bRecover):
                logging.warning('>>> No config file found')
                return

            logging.warning('ConfigLoader::run() - Exception: {}'.format(e))
            logging.warning('>>> Check if backup exists')

            try:
                copyfile( (self.home_path / 'current_config.json.old'), (self.home_path / 'current_config.json'))
            except Exception as e:
                logging.warning('Exception restoring backlup: {}'.format(e))
                pass

            self.loadConfig(bRecover=True)



import time, queue
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):


        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################


        cmd = None
        cnt = 0

        #####################################
        #                                   #
        #     Start of the infinite loop    #
        #                                   #
        #####################################

        # The example executes an infinite loop till it's receives a stop command
        while(True):

            # Example code: Do something
            cnt+=1

            try:
                # Block for 1 second and wait for incoming commands 
                cmd = self.cmd_queue.get(block=True, timeout=1)
            except queue.Empty:
                pass

            if isinstance(cmd, ProcCMD) and cmd.bStop:
                # Stop command received, exit
                return



            # Example Code: Send status text to GUI every timeout interval
            guitext = GuiCMD('cnt: {}'.format(cnt))
            self.return_queue.put(guitext)


            # Example code: Send data to element output every 5 x timeout
            if cnt % 5 == 0:
                # Recors(data, message)
                recordDone = Record(cnt, 'ID: 0x{:08x} - Sending value of cnt: {}'.format(self.id, cnt))     
                self.return_queue.put(recordDone)







import sys, os, itertools, time, termios, tty, select, datetime, logging
from pathlib import Path
from PySide2.QtCore import QThread, Signal


try:
    from screen import reset_screen, reset_screen_dbg
except ImportError: 
    from Pythonic.screen import reset_screen, reset_screen_dbg

class stdinReader(QThread):

    quit_app = Signal()
    finished = Signal()
    b_init      = True
    b_exit      = False
    b_log       = False
    b_procs     = False
    interval    = 0.5
    max_log_lines = 20
    spinner = itertools.cycle(['-', '\\', '|', '/'])

    def __init__(self, refProcessList, logDateStr):
        super().__init__()
        self.proc_list      = refProcessList
        self.startTime      = time.time()
        self.log_date_str   = logDateStr

        # Prepare console
        self.fd = sys.stdin.fileno()
        if os.isatty(sys.stdin.fileno()):
            self.orig_tty_settings = termios.tcgetattr(self.fd) 

    def run(self):

        if self.b_init:
            self.b_init = False
            self.fd = sys.stdin.fileno() 
            if os.isatty(sys.stdin.fileno()):
                self.old_settings = termios.tcgetattr(self.fd) 
                tty.setraw(sys.stdin.fileno()) 

        while not self.b_exit:
            
            rd_fs, wrt_fs, err_fs =  select.select([sys.stdin], [], [], self.interval)

            if rd_fs and os.isatty(sys.stdin.fileno()):
                cmd = rd_fs[0].read(1)

                if cmd == ('q' or 'Q'): # quit
                    termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)
                    termios.tcflush(self.fd, termios.TCIOFLUSH)
                    self.b_exit = True
                    self.quit_app.emit()

                elif cmd == ('p' or 'P'): # show proccesses
                    self.b_procs = True
                    termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)
                    self.printProcessList()
                    tty.setraw(sys.stdin.fileno()) 

                elif cmd == ('l' or 'L'): # show log
                    if self.b_log:
                        termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)
                        reset_screen() # reset the screen to hide the log list
                        reset_screen_dbg()
                        tty.setraw(sys.stdin.fileno()) 
                    self.b_log = not self.b_log
                    
                else:
                    sys.stdout.write('\b')

            else:
                if os.isatty(sys.stdin.fileno()):
                    self.callback()
            


    def callback(self):

        if self.b_procs:
            termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)
            reset_screen() # reset the screen to hide the log list
            reset_screen_dbg()
            tty.setraw(sys.stdin.fileno()) 

        if self.b_log:
            termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)
            reset_screen()
            reset_screen_dbg()
            print('Log output active:\n')
            self.tail(self.max_log_lines)
            tty.setraw(sys.stdin.fileno()) 

        uptime  = time.time() - self.startTime
        minutes = int(uptime // 60 % 60)
        hours   = int(uptime // 3600 % 24)
        days    = int(uptime // 86400)

        sys.stdout.write('Running... ' + next(self.spinner) + 
        '                                           ' +
         'Uptime: {:02d}:{:02d} - {:03d} days'.format(hours, minutes, days))

        sys.stdout.flush()
        sys.stdout.write('\r')



    def tail(self, lines):

        now = datetime.datetime.now().date()
        month = now.strftime('%b')
        year = now.strftime('%Y')
        home_dict = str(Path.home())
        file_path = '{}/Pythonic/log/{}.txt'.format(home_dict, self.log_date_str) 

        BLOCK_SIZE = 1024


        with open(file_path, 'rb') as f:
            f.seek(0, 2) # set fp to 0 from end of file
            block_end_byte = f.tell() # tell() returns the current fp position
            block_number = -1
            blocks = []
            lines_to_go = lines

            while lines_to_go > 0 and block_end_byte > 0:
                if (block_end_byte - BLOCK_SIZE > 0): # bytes to read > BLOCK_SIZE
                    f.seek(block_number * BLOCK_SIZE, 2) # set fp 1 block backwards
                    blocks.append(f.read(BLOCK_SIZE))
                else:
                    f.seek(0,0) # set fp to the beginning
                    blocks.append(f.read(block_end_byte)) # read the rest

                lines_found = blocks[-1].count(b'\n') # count occurences of \n
                lines_to_go -= lines_found
                block_end_byte -= BLOCK_SIZE # move local pointer backwards
                block_number -= 1

            log_display_txt = b''.join(reversed(blocks))
            log_display_txt = b'\n'.join(log_display_txt.splitlines()[-lines:])
            log_display_txt = log_display_txt.decode('utf-8')

            print(log_display_txt + '\n')

    def updateLogDate(self, log_date_str):
        logging.debug('stdinReader::updateLogDate() called with: {}'.format(log_date_str))
        self.log_date_str = log_date_str

    def printProcessList(self):

        # Unix Only
        termios.tcsetattr(self.fd, termios.TCSADRAIN, self.orig_tty_settings)
        reset_screen()
        reset_screen_dbg()
        
        for threadIdentifier, processHandle in self.proc_list:   
            if processHandle.pid:
                print('{} - process, pid: {}'.format(threadIdentifier, processHandle.pid))
            else:
                print('{} - thread'.format(threadIdentifier))

        print('\n')

        tty.setraw(sys.stdin.fileno()) 


import time, itertools, queue
from datetime import datetime, date, time, timedelta
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):

        #interval_str, interval_index, offset, log_state = self.config

        mode             = ''
        self.timebase    = ''
        self.startTime   = ''
        self.endTime     = ''
        self.interval    = 0
        self.tick        = 1 # Threads wait full seconds
        self.activeDays  = []

        recordDone  = Record(None, None)

        specificConfig = self.config.get('SpecificConfig')

        # Set default mode if SpecificConfig is not defined
        # This is the case if the element was created on the working area
        # but the configuration was never opened

        if not specificConfig:

            recordDone = Record(None, message='Trigger: {:04d}'.format(self.config['Identifier']))
            self.return_queue.put(recordDone)
            return

        for attrs in self.config['SpecificConfig']:
            if attrs['Name'] == 'Mode':
                mode = attrs['Data']
            elif attrs['Name'] == 'Timebase':
                self.timebase = attrs['Data']
            elif attrs['Name'] == 'Intervalinput':
                self.interval = attrs['Data']
            elif attrs['Name'] == 'StartTime':
                startTime = attrs['Data']
            elif attrs['Name'] == 'EndTime':
                endTime = attrs['Data']
            elif attrs['Name'] == 'SpecificTime':
                specTime = attrs['Data']
            elif attrs['Name'] == 'Monday'and attrs['Data']:
                self.activeDays.append(0)
            elif attrs['Name'] == 'Tuesday'and attrs['Data']:
                self.activeDays.append(1)
            elif attrs['Name'] == 'Wednesday'and attrs['Data']:
                self.activeDays.append(2)
            elif attrs['Name'] == 'Thursday'and attrs['Data']:
                self.activeDays.append(3)
            elif attrs['Name'] == 'Friday' and attrs['Data']:
                self.activeDays.append(4)
            elif attrs['Name'] == 'Saturday' and attrs['Data']:
                self.activeDays.append(5)
            elif attrs['Name'] == 'Sunday' and attrs['Data']:
                self.activeDays.append(6)


        # Setup interval

        if self.timebase == 'Seconds':
            self.interval = int(self.interval)
            self.tick = 0.2
        elif self.timebase == 'Minutes':
            self.interval = int(self.interval) * 60
        elif self.timebase == 'Hours':
            self.interval = int(self.interval) * 3600

        # Setup start- and endtime

        self.startTime = datetime.strptime(startTime, '%H:%M').time()
        self.stopTime  = datetime.strptime(endTime, '%H:%M').time()
        self.specTime  = datetime.strptime(specTime, '%H:%M').time()

        # Switch modes

        if mode == "Delay":

            self.delayed()
            return

        elif mode == "Interval":

            self.intervalScheduler()
            return

        elif mode == "Interval between times":

            ############################
            #  Interval between times  #
            ############################

            self.intervalBetweenTimes()
            return


        elif mode == "At specific time":

            ############################
            #     At specific time     #
            ############################

            self.atSpecificTime()
            return

        elif mode == "On every full interval":

            ############################
            #  On every full interval  #
            ############################
            self.onEveryFullInterval()
            return

        elif mode == "Full interval between times":

            #################################
            #  Full interval between times  #
            #################################
            self.onEveryFullIntervalbetweenTimes()
            return


    def delayed(self):

        ############################
        #   Single fire, delayed   #
        ############################

        countdown = self.interval / self.tick

        while countdown > 0:


            guitext = GuiCMD(self.remainingTime(countdown=countdown))
            self.return_queue.put(guitext)

            bExit = self.blockAndWait()

            if bExit:
                return

            countdown -= 1

        recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))   
        self.return_queue.put(recordDone)    

    def intervalScheduler(self):

        countdown = self.interval / self.tick

        while True :


            bExit = self.blockAndWait()

            if bExit:
                return

            countdown -= 1

            if countdown <= 0:
                countdown = self.interval / self.tick
                recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))   
                self.return_queue.put(recordDone)
            else:

                # calculate remaining time
                guitext = GuiCMD(self.remainingTime(countdown=countdown))
                self.return_queue.put(guitext)

    def intervalBetweenTimes(self):

        ############################
        #  Interval between times  #
        ############################

        if not self.activeDays:
            return

        nState = 0
        
        while True:
            
            if nState == 0:     # Init: Get the day offset 
                
                dayOffset = self.getDayOffset(self.activeDays, self.stopTime)
                
                if dayOffset == 0 and datetime.now().time() >= self.startTime:
                    nState = 2 # Go to interval mode
                else:
                    nState = 1

                continue

            elif nState == 1:   # Init: Calculate timedelta
                
                delta_t     = datetime.combine(date.today(), self.startTime) - datetime.now()
                delta_t     = delta_t + timedelta(days=dayOffset)      
                nState      = 2
                continue

            elif nState == 2: # Init: Prepare countdown and tick

                countdown   = delta_t.seconds + (delta_t.days * 86400)
                self.tick   = 1
                nState      = 3
                continue

            elif nState == 3:   # Wait for the start
                
                countdown -= 1

                if countdown <= 0:

                    recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))   
                    self.return_queue.put(recordDone)
                    nState = 4 # Go to interval mode
                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext)
                    
            elif nState == 4: # Init Interval Mode

                if self.timebase == 'Seconds':
                    self.tick = 0.2

                countdown = self.interval / self.tick

                nState = 5
                continue

            elif nState == 5: # Interval Mode

                countdown -= 1

                if countdown <= 0:
                    countdown = self.interval / self.tick
                    recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))   
                    self.return_queue.put(recordDone)
                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext)
                
                if datetime.now().time() >= self.stopTime:
                    
                    nState = 0
                    continue


            bExit = self.blockAndWait()

            if bExit:
                return

    def atSpecificTime(self):

        ############################
        #     At specific time     #
        ############################

        if not self.activeDays:
            return

        nState = 0
        
        while True:
            
            if nState == 0:     # Init: Get the day offset 
                
                dayOffset = self.getDayOffset(self.activeDays, self.specTime)
                
                nState = 1
                continue


            elif nState == 1:   # Init: Calculate timedelta
                
                delta_t     = datetime.combine(date.today(), self.specTime) - datetime.now()
                delta_t     = delta_t + timedelta(days=dayOffset)      
                nState      = 2
                continue

            elif nState == 2: # Init: Prepare countdown and tick

                countdown   = delta_t.seconds + (delta_t.days * 86400)
                self.tick   = 1
                nState      = 3
                continue

            elif nState == 3:   # Wait for the start
                            

                if countdown <= 0:

                    recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))   
                    self.return_queue.put(recordDone)
                    nState = 0 # Go to interval mode
                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext)

                countdown -= 1

            bExit = self.blockAndWait()

            if bExit:
                return

    def onEveryFullInterval(self):
        
        ############################
        #  On every full interval  #
        ############################

        nState = 0

        if self.timebase == 'Seconds':
            nState = 10
            self.tick = 0.2
            # Helper value: Prevents that trigger is fired several times when
            # countdown in decrement and the modulo condition is still valid
            # Init with an 'invalid' value (according to the timebase)
            lastFired = 61 
        elif self.timebase == 'Minutes':
            nState = 20
        elif self.timebase == 'Hours':
            nState = 30


        # Countdown muss korrekt initialisiert werden
        countdown = self.interval / self.tick        

        while True:
        
            countdown   -= 1
            time        = datetime.now().time()

            if nState == 10:     # Every full second: Init countdown


                # passt
                #x = (self.interval / self.tick)
                #y = ((second % self.interval) / self.tick)
                countdown -= (time.second % self.interval) / self.tick
                
                nState = 11
                continue
        
            elif nState == 11:    # Every full second

                #countdown -= 1

                if  time.second % self.interval == 0 and lastFired != time.second:
                    recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))    
                    self.return_queue.put(recordDone)
                    countdown = self.interval / self.tick
                    lastFired = time.second

                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext)

            elif nState == 20: # Every full minutes: Init countdown

                
                
                # Calculate minutes
                fullMinutesInterval     = self.interval // 60
                passedMinutes           = time.minute % fullMinutesInterval
                countdown               -= (passedMinutes * 60 ) / self.tick

                # Calculate seconds
                countdown               -= time.second / self.tick            

                nState = 21
                continue


            elif nState == 21: # Every full minutes


                if  time.minute % (self.interval / 60) == 0 and time.second == 0:
                    recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))    
                    self.return_queue.put(recordDone)
                    countdown = self.interval / self.tick

                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext)


            elif nState == 30: # Every full hours: Init countdown
                
                
                # Calculate hours
                fullHoursInterval       = self.interval // 3600
                passedHours             = time.hour % fullHoursInterval
                countdown               -= (passedHours * 3600 )/ self.tick

                # Calculate minutes
                fullMinutesInterval     = self.interval // 60
                passedMinutes           = time.minute % fullMinutesInterval
                countdown               -= (passedMinutes * 60 )/ self.tick

                # Calculate seconds
                countdown               -= time.second / self.tick            

                nState = 21
                continue


            elif nState == 31: # Every full hours


                if  time.minute % (self.interval / 60) == 0 and time.second == 0:
                    recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))    
                    self.return_queue.put(recordDone)
                    countdown = self.interval / self.tick

                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext) 

            bExit = self.blockAndWait()

            if bExit:
                return

    def onEveryFullIntervalbetweenTimes(self):

        #########################################
        #  On every fullinterval between times  #
        #########################################

        if not self.activeDays:
            return

        if self.timebase == 'Seconds':
            self.tick = 0.2

        countdown = self.interval / self.tick  

        nState = 0
        
        while True:

            time = datetime.now().time()
            
            if nState == 0:     # Init: Get the day offset 
                
                dayOffset = self.getDayOffset(self.activeDays, self.stopTime)
                
                if dayOffset == 0 and time >= self.startTime:
                    nState = 2 # Go to interval mode
                else:
                    nState = 1

                continue

            elif nState == 1:   # Init: Calculate timedelta
                
                delta_t     = datetime.combine(date.today(), self.startTime) - datetime.now()
                delta_t     = delta_t + timedelta(days=dayOffset)      
                nState      = 2
                continue

            elif nState == 2: # Init: Prepare countdown and tick

                countdown   = delta_t.seconds + (delta_t.days * 86400)
                self.tick   = 1
                nState      = 3
                continue

            elif nState == 3:   # Wait for the start
                
                countdown -= 1

                if countdown <= 0:
                    nState = 4 # Go to interval mode
                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext)
                    
            elif nState == 4: # Init Interval Mode

                if self.timebase == 'Seconds':
                    nState = 50
                    # Helper value: Prevents that trigger is fired several times when
                    # countdown in decrement and the modulo condition is still valid
                    # Init with an 'invalid' value (according to the timebase)
                    lastFired = 61 
                elif self.timebase == 'Minutes':
                    nState = 60
                elif self.timebase == 'Hours':
                    nState = 70

                continue

            elif nState == 50: # Every full second: Init countdown
                  

                countdown -= (time.second % self.interval) / self.tick
                
                nState = 51
                continue
                
            elif nState == 51:    # Every full second

                countdown   -= 1


                if  time.second % self.interval == 0 and lastFired != time.second:
                    recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))    
                    self.return_queue.put(recordDone)
                    countdown = self.interval / self.tick
                    lastFired = time.second

                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext)

                if time >= self.stopTime:
                    
                    nState = 0
                    continue

            elif nState == 60: # Every full minutes: Init countdown
            
                # Calculate minutes
                fullMinutesInterval     = self.interval // 60
                passedMinutes           = time.minute % fullMinutesInterval
                countdown               -= (passedMinutes * 60 ) / self.tick

                # Calculate seconds
                countdown               -= time.second / self.tick            

                nState = 61
                continue


            elif nState == 61: # Every full minutes

                countdown -= 1

                if  time.minute % (self.interval / 60) == 0 and time.second == 0:
                    recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))    
                    self.return_queue.put(recordDone)
                    countdown = self.interval / self.tick

                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext)

                if time >= self.stopTime:
                    
                    nState = 0
                    continue


            elif nState == 70: # Every full hours: Init countdown
                        
                        
                # Calculate hours
                fullHoursInterval       = self.interval // 3600
                passedHours             = time.hour % fullHoursInterval
                countdown               -= (passedHours * 3600 )/ self.tick

                # Calculate minutes
                fullMinutesInterval     = self.interval // 60
                passedMinutes           = time.minute % fullMinutesInterval
                countdown               -= (passedMinutes * 60 )/ self.tick

                # Calculate seconds
                countdown               -= time.second / self.tick            

                nState = 71
                continue


            elif nState == 71: # Every full hours

                countdown -= 1

                if  time.minute % (self.interval / 60) == 0 and time.second == 0:
                    recordDone = Record(data=None, message='>Trigger<   ProcHdl::{:04d}'.format(self.config['Identifier']))  
                    self.return_queue.put(recordDone)
                    countdown = self.interval / self.tick

                else:

                    # calculate remaining time
                    guitext = GuiCMD(self.remainingTime(countdown=countdown))
                    self.return_queue.put(guitext) 

                if time >= self.stopTime:
                    
                    nState = 0
                    continue


            bExit = self.blockAndWait()

            if bExit:
                return


    def blockAndWait(self):

        cmd = None
        
        try:
            # Wait for incoming commands in specified interval
            cmd = self.cmd_queue.get(block=True, timeout=self.tick)
        except queue.Empty:
            pass

        if isinstance(cmd, ProcCMD) and cmd.bStop:
            # Exit here if stop command received
            return True

        return False


    def remainingTime(self, countdown=None):

        startDay = date.today()
   
        
        if not countdown:
            delta_t = datetime.combine(startDay, self.startTime) - datetime.now()
        else:
            delta_t = timedelta(seconds=countdown*self.tick)

        #delta_t = self.chop_microseconds(delta_t)
        hours           = delta_t.seconds // 3600
        minutes         = (delta_t.seconds // 60) % 60
        seconds         = delta_t.seconds % 60
        milliseconds    = delta_t.microseconds // 100000
        ##sTimeDelta = str(timedelta)

        if delta_t.seconds < 10: # return milliseconds   
            return '{:02d}.{}'.format(seconds, milliseconds)
        elif delta_t.seconds < 60 and delta_t.days == 0: # return full seconds
            return '{:02d}:{:02d}'.format(minutes, seconds)
        elif delta_t.days == 0: # return hh:mm:ss
            return '{:02d}:{:02d}:{:02d}'.format(hours, minutes, seconds)
        else: #return dd hh:mm:ss
            return '{} {:02d}:{:02d}:{:02d}'.format(delta_t.days, hours, minutes, seconds)
            


    def getDayOffset(self, activeDays, stopTime):

        #########################################
        #                                       #
        #        Calculate the start day        #
        #                                       #
        #########################################

        now             = datetime.now()
        delta_t_days    = 0
        currentWeekday  = now.weekday()
        iterDays        = itertools.cycle(activeDays) 
        start_day       = next(iterDays)
        
        # start day == today or later that week     
        if any(i >= currentWeekday for i in activeDays):
            
            while start_day < currentWeekday:
                start_day = next(iterDays)

            delta_t_days = start_day - currentWeekday

        # start day is next week
        else:
            delta_t_days = 7 - currentWeekday + start_day
        

        # Check if the start time already passed

        if start_day == currentWeekday and stopTime <= now.time():             
            start_day =next(iterDays)
            if start_day <= currentWeekday:
                # rest of the week + start day
                delta_t_days = 7 - currentWeekday + start_day
            else:
                delta_t_days = start_day - currentWeekday

        return delta_t_days



    def chop_microseconds(self, delta):
        return delta - timedelta(microseconds=delta.microseconds)



import time, queue
import ccxt
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):


        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################


        specificConfig = self.config.get('SpecificConfig')

        if not specificConfig:

            recordDone = Record(None, message='Trigger: {:04d}'.format(self.config['Identifier']))
            self.return_queue.put(recordDone)
            return


        eId     = None
        pubKey  = None
        prvKey  = None


        for attrs in specificConfig:
            if attrs['Name'] == 'ExchangeId':
                eId = attrs['Data']
            if attrs['Name'] == 'PubKey':
                pubKey = attrs['Data']
            elif attrs['Name'] == 'PrvKey':
                prvKey = attrs['Data']


        exchangeClass = getattr(ccxt, eId)
        if pubKey and prvKey:

             exchange = exchangeClass( {
                                        'apiKey'            : pubKey,
                                        'secret'            : prvKey,
                                        'enableRateLimit'   : True
             })

        else:
            exchange = exchangeClass( {'enableRateLimit'   : True})

        method = getattr(exchange, self.inputData['method'])

        kwargs  = None
        params  = None
        
        if not 'kwargs' in self.inputData:

            data = method()

        elif not 'params' in self.inputData:
            
            kwargs = self.inputData['kwargs']
            data = method(**kwargs)

        else: 

            kwargs = self.inputData['kwargs']
            params = self.inputData['params']

            if params != '':
                data = method(**kwargs, params=params)
            else:
                data = method(**kwargs)

           
        recordDone = Record(data, '{}() successfull'.format(self.inputData['method']))     
        self.return_queue.put(recordDone)



        


import eventlet, os, sys, json
import debugpy
from pathlib import Path
from web_daemon import MainWorker
from PySide2.QtCore import QCoreApplication, QTimer

os.environ['PYTHONWARNINGS'] = 'ignore:semaphore_tracker:UserWarning'



def run():
    
    # Create home path (if not already existing)
    
    home_path = Path.home() / 'Pythonic'
    home_path.mkdir(exist_ok=True)

    # Create log path (if not already existing)

    logPath = home_path / 'log'
    logPath.mkdir(exist_ok=True)

    # Create directory for executables (if not already existing)

    execPath = home_path / 'executables'
    execPath.mkdir(exist_ok=True)

    # Create directory for vs code configuration (if not already existing)

    vsCodepath = execPath / '.vscode' 
    vsCodepath.mkdir(exist_ok=True)

    # Append executables folder to module search path
    sys.path.append(str(execPath))


    """
    Create launch.json with current PID in hte libraries installation path
    """
    libPath = os.path.dirname(__file__)

    Path(os.path.join(libPath, '.vscode/')).mkdir(exist_ok=True)

    launch = {  "version": "0.2.0",
                "configurations": [
            {
                "name": "Pythonic: Attach",
                "type": "python",
                "request": "attach",
                "justMyCode": False,
                "connect" : {
                    "host" : "localhost",
                    "port"  : 5678
                }
            }
        ]
    }



    """
    Create launch.json with current PID in the user executables path
    """
    
    with open(vsCodepath / 'launch.json', 'w') as file:
        json.dump(launch, file, indent=4)
    



    app = QCoreApplication(sys.argv)
    ex = MainWorker(app)
    ex.start(sys.argv)
    
    app.exec_()

if __name__ == '__main__':

    run()



import time, queue, pickle, ssl, smtplib
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD

from email.message import EmailMessage
    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):


        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################
        specificConfig = self.config.get('SpecificConfig')

        if not specificConfig:

            recordDone = Record(None, message='Config missing')
            self.return_queue.put(recordDone)
            return
        
        sender      = None
        password    = None
        url         = None
        port        = None
        
        recipients  = None
        subject     = None
        message     = None

        attachments = None


        for attrs in specificConfig:
            if attrs['Name'] == 'Sender':
                sender = attrs['Data']
            elif attrs['Name'] == 'Password':
                password = attrs['Data']
            elif attrs['Name'] == 'URL':
                url = attrs['Data']
            elif attrs['Name'] == 'Port':
                port = int(attrs['Data'])

        if not sender:
            raise Exception('Sender missing in configuration')
        if not password:
            raise Exception('Password missing in configuration')
        if not url:
            raise Exception('URL missing in configuration')
        if not port:
            raise Exception('Port missing in configuration')


        if isinstance(self.inputData, dict):
            if not 'recipient' in self.inputData or not isinstance(self.inputData['recipient'], str):
                recordDone = Record(PythonicError('Key error, see log for details'), message='Key "recipient" not found or not of type string')
                self.return_queue.put(recordDone)
                return
            if not 'subject' in self.inputData or not isinstance(self.inputData['subject'], str):
                recordDone = Record(PythonicError('Key error, see log for details'), message='Key "subject" not found or not of type string')
                self.return_queue.put(recordDone)
                return
            if not 'message' in self.inputData or not isinstance(self.inputData['message'], str):
                recordDone = Record(PythonicError('Key error, see log for details'), message='Key "subject" not found or not of type string')
                self.return_queue.put(recordDone)
                return

            recipients  = self.inputData['recipient']
            subject     = self.inputData['subject']
            message     = self.inputData['message']

            # optional: check for attachment(s)
            if 'attachment' in self.inputData and isinstance(self.inputData['attachment'], list):

                attachments = self.inputData['attachment']

        else:
            recordDone = Record(PythonicError('Config missing'), message='Config missing')
            self.return_queue.put(recordDone)
            return

        msg = EmailMessage()
        msg['Subject']  = subject
        msg['From']     = sender
        msg['To']       = recipients
        msg.set_default_type('text/plain')
        msg.set_content(message)

        if attachments:
            for attachment in attachments:
                if not 'filename' in attachment: # and not isinstance(attachment['filename'], str):
                    continue
                if not isinstance(attachment['filename'], str):
                    continue
                if not 'data' in attachment:
                    continue

                # attach data as text
                if isinstance(attachment['data'], str):
                    msg.add_attachment(attachment['data'], 'text/plain', filename=attachment['filename'])

                else: # attach data is binary object
                    msg.add_attachment(pickle.dumps(attachment['data']), maintype='application', subtype='octet-stream',
                        filename=attachment['filename'])



        context = ssl.create_default_context()

        with smtplib.SMTP_SSL(url, port, context=context) as server:
            server.login(sender, password)
            server.send_message(msg)


import logging, pickle
from pathlib import Path
from typing import Optional

class Function():

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        #logging.debug('Function.__init__()')
        self.id             = id
        self.config         = config
        self.inputData      = inputData
        self.return_queue   = return_queue
        self.cmd_queue      = cmd_queue
        self.logger         = logging.getLogger() # mt only
        self.bStop          = False

        self.logger.setLevel(logging.DEBUG)

    def __setstate__(self, state):
        #logging.debug('__setstate__() called Function')
        self.id, self.config, self.inputData, self.return_queue, self.cmd_queue, self.logger, self.bStop = state

    def __getstate__(self):
        #logging.debug('__getstate__() called Function')
        return (self.id, self.config, self.inputData, self.return_queue, self.cmd_queue, self.logger, self.bStop)

    def execute_ex(self):

        #logging.debug('execute_ex() called Function')

        try:
            result = self.execute()
        except Exception as e:
            guiexc = GuiException(e)
            self.return_queue.put(guiexc)


class Record():

    def __init__(self, data : any, message : Optional[str] = None) -> None:

        self.data       = data # Will be passed to subsequent elements, even it is None
        self.message    = message # Log message string


    def __setstate__(self, state):
        #logging.debug('__setstate__() called Record')
        self.data, self.message = state

    def __getstate__(self):
        #logging.debug('__getstate__() called Record')
        return(self.data, self.message)


class PythonicError:

    def __init__(self, msg : str) -> None:
        self.msg = msg
    
    def  __str__(self) -> str:
        return 'Error: {}'.format(self.msg)

class GuiCMD:
    
    def __init__(self, text):
        self.text = text

    def __setstate__(self, state):
        self.text = state

    def __getstate__(self):
        return(self.text)

class GuiException:

    def __init__(self, e : Exception) -> None:
        self.e = e

class ProcCMD:

    def __init__(self, data, bStop=False):
        self.bStop  = bStop
        self.data   = data

    def __setstate__(self, state):
        #logging.debug('__setstate__() called Record')
        self.bStop, self.data = state

    def __getstate__(self):
        #logging.debug('__getstate__() called Record')
        return(self.bStop, self.data)


def storePersist(func):

    def wrapper(self, *args, **kwargs):

        func(self, *args, **kwargs)

        with open(self.filename, 'wb') as f:
            try:
                pickle.dump(self.copy(), f)
            except Exception as e:
                pass
        
    return wrapper


class ListPersist(list):

    def __init__(self, name: str) -> None:

        self.filename = Path.home() / 'Pythonic' / 'executables' / '{}.obj'.format(name)

        if self.filename.exists():

            with open(self.filename, 'rb') as f:

                data = pickle.load(f)
                super().extend(data)



    def reload(self):

        if self.filename.exists():

            with open(self.filename, 'rb') as f:

                data = pickle.load(f)
                super().clear()
                super().extend(data)
                return True

        else:
            return False


    

    @storePersist
    def append(self, __object) -> None:
        return super().append(__object)

    @storePersist
    def extend(self, __iterable) -> None:
        return super().extend(__iterable)

    @storePersist
    def remove(self, __value) -> None:
        return super().remove(__value)

    @storePersist
    def pop(self, __index: int):
        return super().pop(__index=__index)

    @storePersist
    def clear(self) -> None:
        return super().clear()


class SetPersist(set):

    def __init__(self, name: str) -> None:

        self.filename = Path.home() / 'Pythonic' / 'executables' / '{}.obj'.format(name)

        if self.filename.exists():

            with open(self.filename, 'rb') as f:

                data = pickle.load(f)
                super().update(data)



    def reload(self):

        if self.filename.exists():

            with open(self.filename, 'rb') as f:

                data = pickle.load(f)
                super().clear()
                super().extend(data)
                return True

        else:
            return False



    @storePersist
    def add(self, __object) -> None:
        return super().add(__object)

    @storePersist
    def discard(self, __object) -> None:
        return super().discard(__object)

    @storePersist
    def pop(self) -> None:
        return super().pop()

    @storePersist
    def clear(self) -> None:
        return super().clear()


import sys, logging, locale, datetime, os, signal, time, json, argparse
import multiprocessing as mp
from eventlet import wsgi, websocket, greenthread, listen
from threading import Timer, Thread, Event
from pathlib import Path
from zipfile import ZipFile
from enum import Enum
from shutil import copyfile
from PySide2.QtCore import QCoreApplication, QObject, QThread, Qt
from PySide2.QtCore import Signal

try:
    from logfile_hanlder import LogFileHandler
    from execution_operator import Operator   
    from screen import reset_screen, reset_screen_dbg
    from configio import ToolboxLoader, ConfigLoader, EditorLoader, ConfigWriter, ExecSysCMD
except ImportError:
    from Pythonic.logfile_hanlder import LogFileHandler
    from Pythonic.execution_operator import Operator
    from Pythonic.screen import reset_screen, reset_screen_dbg
    from Pythonic.configio import ToolboxLoader, ConfigLoader, EditorLoader, ConfigWriter, ExecSysCMD


##############################################
#                                            #
#                GLOBAL PATHS                #
#                                            #
##############################################

www_root    = 'public_html/'
www_static  = 'public_html/static/'
www_config  = 'public_html/config/'
executables = 'executables'


# Allow to quit with CTRL+C
signal.signal(signal.SIGINT, signal.SIG_DFL)


class LogLvl(Enum):
    DEBUG       = 0
    INFO        = 1
    WARNING     = 2
    CRITICAL    = 3
    FATAL       = 4

# Replace the websockets with QWebSockets
@websocket.WebSocketWSGI
def rcv(ws):

    #logging.getLogger()

    def send(command):
        #logging.debug('send() - command: {}'.format(command['cmd']))
        try:
            ws.send(json.dumps(command))
        except Exception as e:
            logging.info('PythonicWeb - RCV Socket connection lost: {}'.format(e))
            ws.environ['mainWorker'].frontendCtrl.disconnect(send)
 
    ws.environ['mainWorker'].frontendCtrl.connect(send)
    
    bConnected = True
    while bConnected:
        greenthread.sleep(0.1)
        QCoreApplication.processEvents()
        greenthread.sleep(0.1)
        QCoreApplication.processEvents()
        greenthread.sleep(0.1)
        QCoreApplication.processEvents()

        try:

            date = datetime.datetime.now() #.strftime("%d-%b-%Y")
            jsonHeartBeat = {   'cmd'       : 'Heartbeat',
                                'address'   : { 'target' : 'MainWindow'} ,
                                'data' : date.strftime("%d-%b-%Y %H:%M:%S") }

            ws.send(json.dumps(jsonHeartBeat))

        except Exception as e:
            logging.info('PythonicWeb - RCV Socket connection lost: {}'.format(e))
            ws.environ['mainWorker'].frontendCtrl.disconnect(send)
            bConnected = False

    logging.debug('PythonicWeb - RCV Socket closed')


@websocket.WebSocketWSGI
def ctrl(ws):
    logging.debug('PythonicDaemon - CTRL WebSocket connected')   
    while True:
        m = ws.wait()
        if m is None:
            logging.debug('PythonicDaemon - CTRL Socket Closed')            
            break;   
        else:

            msg = json.loads(m)
            
            logging.debug('PythonicWeb    - Command: {}'.format(msg['cmd']))

            if msg['cmd'] == 'logMsg':
                # logging
                logObj = msg['data']

                if logObj['logLvL'] == LogLvl.DEBUG.value:
                    logging.debug('PythonicWeb    - {}'.format(logObj['msg']))
                if logObj['logLvL'] == LogLvl.DEBUG.value:
                    logging.debug('PythonicWeb    - {}'.format(logObj['msg']))
                elif logObj['logLvL'] == LogLvl.INFO.value:
                    logging.info('PythonicWeb    - {}'.format(logObj['msg']))
                elif logObj['logLvL'] == LogLvl.WARNING.value:
                    logging.warning('PythonicWeb    - {}'.format(logObj['msg']))
                elif logObj['logLvL'] == LogLvl.CRITICAL.value:
                    logging.error('PythonicWeb    - {}'.format(logObj['msg']))
                elif logObj['logLvL'] == LogLvl.FATAL.value:
                    logging.critical('PythonicWeb    - {}q'.format(logObj['msg']))


            elif msg['cmd'] == 'writeConfig':
                logging.debug('Config loaded')
                # Save config to to memory
                ws.environ['mainWorker'].config = msg['data']
                # Save config to file
                ws.environ['mainWorker'].saveConfig.emit(msg['data'])
            elif msg['cmd'] == 'StartExec':
                #logging.info(h.heap())
                elementId = msg['data']      
                ws.environ['mainWorker'].startExec.emit(elementId, ws.environ['mainWorker'].config)
            elif msg['cmd'] == 'StopExec':
                elementId = msg['data']
                ws.environ['mainWorker'].stopExec.emit(elementId)
                logging.debug('PythonicWeb    - {}'.format(msg['cmd']))
            elif msg['cmd'] == 'QueryConfig':
                logging.debug('PythonicWeb    - {}'.format(msg['cmd']))
                ws.environ['mainWorker'].loadConfig()
            elif msg['cmd'] == 'QueryToolbox':
                logging.debug('PythonicWeb    - {}'.format(msg['cmd']))
                ws.environ['mainWorker'].loadTools()
            elif msg['cmd'] == 'QueryEditorToolbox' :
                logging.debug('PythonicWeb    - {}'.format(msg['cmd']))
                addr        = msg['address']
                typeName    = msg['data']
                ws.environ['mainWorker'].loadEditorConfig(addr, typeName)
            elif msg['cmd'] == 'SysCMD' :
                logging.debug('PythonicWeb    - {}'.format(msg['cmd']))
                ws.environ['mainWorker'].sysCommand.emit(msg['data'])
            elif msg['cmd'] == 'QueryElementStates' :
                logging.debug('PythonicWeb    - {}'.format(msg['cmd']))
                ws.environ['mainWorker'].queryStates.emit()
            elif msg['cmd'] == 'StartAll' :
                logging.debug('PythonicWeb    - {}'.format(msg['cmd']))
                ws.environ['mainWorker'].startAll.emit(ws.environ['mainWorker'].config)
            elif msg['cmd'] == 'StopAll' :
                logging.debug('PythonicWeb    - {}'.format(msg['cmd']))
                ws.environ['mainWorker'].stopAll.emit()
            elif msg['cmd'] == 'KillAll' :
                logging.debug('PythonicWeb    - {}'.format(msg['cmd']))
                ws.environ['mainWorker'].killAll.emit()
        
        del msg
        del m
             
@websocket.WebSocketWSGI
def saveConfig(ws):
    filename = ws.wait()
    logging.info('Upload Config: Filename: {}'.format(filename))
    data = ws.wait()

    # Create a backup of existing config
    home_path = Path.home() / 'Pythonic'
    try:
        copyfile( (home_path / 'current_config.json'), (home_path / 'current_config.json.old'))
    except Exception as e:
        logging.warning('Exception during creating backlup: {}'.format(e))
        pass
    
    new_file = os.path.join(www_config, 'current_config.json')
    data_size = float(len(data)) / 1000 #kb
    logging.info('Sizeof Config: {:.1f} kb'.format(data_size))
    # Backup previous config

    logging.info('Upload saved to: {}'.format(home_path / 'current_config.json'))
    with open((home_path / 'current_config.json'), 'wb') as file:
        file.write(data)
    
    ws.environ['mainWorker'].loadConfig()

@websocket.WebSocketWSGI
def saveExecutable(ws):
    filename = ws.wait()
    logging.info('Upload Executable: Filename: {}'.format(filename))
    data = ws.wait()
    data_size = float(len(data)) / 1000 #kb
    logging.info('Sizeof Config: {:.1f} kb'.format(data_size))
    home_path = Path.home() / 'Pythonic'

    #new_file = os.path.join(executables, filename)
    logging.info('Upload saved to: {}'.format(home_path / executables / filename))


    with open(home_path / executables /filename, 'wb') as file:
        file.write(data)

def dispatch(environ, start_response):

    """
        WEBSOCKETS
    """

    png_req4 = environ['PATH_INFO'][-4:] # last 4 characters '.png'
    png_req3 = environ['PATH_INFO'][-3:] # last 4 characters '.js'

    if environ['PATH_INFO'] == '/ctrl':
        logging.debug('PythonicDaemon - Open CTRL WebSocket')     
        return ctrl(environ, start_response)
    elif environ['PATH_INFO'] == '/rcv':
        logging.debug('PythonicDaemon - Open RCV WebSocket')     
        return rcv(environ, start_response)

    elif environ['PATH_INFO'] == '/config':
        logging.debug('PythonicDaemon - Open Config WebSocket')  
        return saveConfig(environ, start_response)

    elif environ['PATH_INFO'] == '/executable':
        logging.debug('PythonicDaemon - Open Config WebSocket')  
        return saveExecutable(environ, start_response)

    elif environ['PATH_INFO'] == '/log':
        logging.debug('PythonicDaemon - Providing list of log files')  
        log_files = os.listdir(Path.home() / 'Pythonic' / 'log')
        log_files = '\n'.join(log_files)

        start_response('200 OK', [  ('content-type', 'text/plain; charset=utf-8') ])                              
        
        return [log_files]


        ###########################
        # STANDARD HTML ENDPOINTS #
        ###########################

    elif environ['PATH_INFO'] == '/':
        #logging.debug('PATH_INFO == \'/\'')
        
        start_response('200 OK', [  ('content-type', 'text/html'),
                                    ('Cross-Origin-Opener-Policy', 'same-origin'),
                                    ('Cross-Origin-Embedder-Policy', 'require-corp')])
        return [open(os.path.join(os.path.dirname(__file__),
            www_root + 'templates/PythonicWeb.html')).read()]


    elif environ['PATH_INFO'] == '/qtlogo.svg':
        #logging.debug('PATH_INFO == \'/qtlogo.svg\'')

        open_path = os.path.join(os.path.dirname(__file__), www_static + 'qtlogo.svg')

        with open(open_path,'rb') as f:
            img_data = f.read()

        start_response('200 OK', [('content-type', 'image/svg+xml'),
                                ('content-length', str(len(img_data)))])

        return [img_data]

    elif environ['PATH_INFO'] == '/favicon.ico':
        #logging.debug('PATH_INFO == \'/qtlogo.svg\'')

        open_path = os.path.join(os.path.dirname(__file__), www_static + 'python.ico')

        with open(open_path,'rb') as f:
            img_data = f.read()

        start_response('200 OK', [('content-type', 'image/vnd.microsoft.icon'),
                                ('content-length', str(len(img_data)))])

        return [img_data]
        

    # IMAGES (*.png)
    elif png_req4 == '.png':
        logging.debug('PATH_INFO == ' + environ['PATH_INFO'])
        
        open_path = os.path.join(os.path.dirname(__file__), www_static + environ['PATH_INFO'])

        with open(open_path,'rb') as f:
            img_data = f.read()
        
        start_response('200 OK', [('content-type', 'image/png'),
                                ('content-length', str(len(img_data)))])
        
        return [img_data]

    # JAVS SCRIPT (*.js)

    elif png_req3 == '.js':
        #logging.debug('PATH_INFO == ' + environ['PATH_INFO'])
        open_path = os.path.join(os.path.dirname(__file__), www_static + environ['PATH_INFO'])
        with open(open_path,'rb') as f:
            img_data = f.read()
      
        start_response('200 OK', [('content-type', 'application/javascript')])
        
        return [img_data]

    
    # Executable (*.py)

    elif png_req3 == '.py':
        #logging.debug('PATH_INFO == ' + environ['PATH_INFO'])
        open_path = Path.home() / 'Pythonic' / 'executables' / environ['PATH_INFO'][1:]
        with open(open_path,'rb') as f:
            py_file = f.read()
      
        start_response('200 OK', [('content-type', 'application/javascript')])
        
        return [py_file]
    

    elif environ['PATH_INFO'] == '/PythonicWeb.wasm':
        #logging.debug('PATH_INFO == \'/PythonicWeb.wasm\'')

        open_path = os.path.join(os.path.dirname(__file__), www_static + 'PythonicWeb.wasm')

        with open(open_path,'rb') as f:
            bin_data = f.read()

        start_response('200 OK', [('content-type', 'application/wasm')])
        return [bin_data]	

    elif environ['PATH_INFO'] == '/PythonicWeb.data':
        #logging.debug('PATH_INFO == \'/PythonicWeb.data\'')

        open_path = os.path.join(os.path.dirname(__file__), www_static + 'PythonicWeb.data')

        with open(open_path,'rb') as f:
            bin_data = f.read()

        start_response('200 OK', [('content-type', 'application/wasm')])
        return [bin_data]	

    elif png_req4 == '.txt': # Log files
        #logging.debug('PATH_INFO == \'/PythonicWeb.wasm\'')

        open_path = Path.home() / 'Pythonic' / 'log' / environ['PATH_INFO'][1:]
        
        with open(open_path,'rb') as f:
            log_data = f.read()

        start_response('200 OK', [  ('content-type', 'text/plain; charset=utf-8')])                                                                   
                                                 
        return [log_data]

    # Config file (*.json)

    elif environ['PATH_INFO'] == '/current_config.json':
        #logging.debug('PATH_INFO == ' + environ['PATH_INFO'])
        open_path = Path.home() / 'Pythonic' / 'current_config.json'
        with open(open_path,'rb') as f:
            config_file = f.read()
      
        start_response('200 OK', [  ('content-type', 'application/json; charset=utf-8')])
        
        return [config_file]

    else:
        path_info = environ['PATH_INFO']
        logging.debug('PATH_INFO = {}'.format(path_info))
        return None


class WSGI_Server(QThread):

    def __init__(self, mainWorker):
        super().__init__()
        self.mainWorker = { 'mainWorker' : mainWorker }

    def run(self):

        #listener = eventlet.listen(('127.0.0.1', 7000))
        listener = listen(('0.0.0.0', 7000))
        wsgi.server(listener, dispatch, log_output=False, environ=self.mainWorker)


class MainWorker(QObject):

    kill_all        = Signal()
    

    max_grid_size   = 50
    max_grid_cnt    = 5
    config          = None # element configuration

    update_logdate  = Signal(object)        # update displayed date string in stdin_reader
    startExec       = Signal(object, object)# element-Id, configuration
    stopExec        = Signal(object)        # element-Id
    saveConfig      = Signal(object)        # configuration
    sysCommand      = Signal(object)        # Optional: Element Constructor / Destructor
    frontendCtrl    = Signal(object)
    queryStates     = Signal()              # Query the running states of elements
    startAll        = Signal(object)        # Start all elements: (config)
    stopAll         = Signal()              # Stop all elements
    killAll         = Signal()              # Kill all running processes

    def __init__(self, app):
        super(MainWorker, self).__init__()
        self.app = app

        # Setup command line arguments
        
        parser = argparse.ArgumentParser(description='Pythonic background daemon')
        # Debug output switch 
        parser.add_argument('-Ex', action='store_true', help='Interactive shell interface, Unix only')
        # Log level
        parser.add_argument('-v', action='store_true', help='Verbose output')

        self.args = parser.parse_args()

        # Set Log Level
        if self.args.v:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO

        # Instantiate the LogFileHandler
        self.logFileHandler = LogFileHandler(log_level)

        # Select multiprocessing spawn method
        mp.set_start_method('spawn')
        
        # Set working directory
        os.chdir(Path(__file__).parent.absolute())

        # Setup format
        locale.setlocale(locale.LC_TIME, '')

        # Instantiate WSGI Server
        self.wsgi_server = WSGI_Server(self)
        
        # Instantiate Execution Operator
        self.operator = Operator()
        self.operator.command.connect(self.forwardCmd)
        self.startExec.connect(self.operator.startExec)
        self.stopExec.connect(self.operator.stopExec)
        self.queryStates.connect(self.operator.getElementStates)
        self.startAll.connect(self.operator.startAll)
        self.stopAll.connect(self.operator.stopAll)
        self.killAll.connect(self.operator.killAll)
        

        # Instantiate Standard Input Reader (Unix only)
        if self.args.Ex:
            try:
                from stdin_reader import stdinReader
            except ImportError:
                from Pythonic.stdin_reader import stdinReader        
        
            self.stdinReader = stdinReader(self.operator.processHandles.items(), self.logFileHandler.log_date_str)
            self.stdinReader.quit_app.connect(self.exitApp)


        # Connect the logger
        if self.args.Ex:
            self.logFileHandler.update_logdate.connect(self.stdinReader.updateLogDate)
               

        # Instantiate ToolboxLoader
        self.toolbox_loader = ToolboxLoader()
        self.toolbox_loader.tooldataLoaded.connect(self.forwardCmd)
        
        # Instantiate (Element)-EditorLoader

        self.editor_loader = EditorLoader()
        self.editor_loader.editorLoaded.connect(self.forwardCmd)
        
        # Instantiate ConfigWriter
        self.config_writer = ConfigWriter()
        self.config_writer.configSaved.connect(self.forwardCmd)
        self.saveConfig.connect(self.config_writer.saveConfig)

        # Instantiate ConfigLoader
        self.config_loader = ConfigLoader()
        self.config_loader.configLoaded.connect(self.configLoaded)

        # Instantiate System Command Executor
        self.exec_sys_cmd = ExecSysCMD()
        self.sysCommand.connect(self.exec_sys_cmd.execCommand)
            
    
    def exitApp(self):
        print('# Stopping all processes....')
        self.kill_all.emit()
        time.sleep(3) # wait for 1 seconds to kill all processes
        self.app.quit()
        os.kill(self.app.applicationPid(), signal.SIGTERM) # kill all related threads
    

    def start(self, args):

        
        reset_screen()    

        if self.args.Ex:
            reset_screen_dbg()    
            self.stdinReader.start() # call run() method in separate thread

        self.config = self.config_loader.loadConfigSync()


        self.wsgi_server.start()
        self.operator.start(self.config)

    def loadTools(self): # Multithreaded
        
        logging.debug('MainWorker::loadTools() called')
        self.toolbox_loader.start()

    def loadEditorConfig(self, address, typeName): # Multithreaded
        
        logging.debug('MainWorker::loadEditorConfig() called')
        self.editor_loader.startLoad(address, typeName)

    def forwardCmd(self, cmd):

        #logging.debug('MainWorker::forwardCmd() called')
        self.frontendCtrl.emit(cmd)

    def loadConfig(self): # Multithreaded
        
        logging.debug('MainWorker::loadConfig() called')
        self.config_loader.start()
           
    def configLoaded(self, config):

        self.config = config

        address = { 'target' : 'MainWindow'}
        
        cmd = { 'cmd'       : 'CurrentConfig',
                'address'   : address,
                'data'      : config }

        self.frontendCtrl.emit(cmd)

    def checkArgs(self, args):

        b_file_found = False
        grid_file = None

        for argument in args:
            if argument[0] == '-' or argument[0] == '--':
                print('Option found: {}'.format(argument))
            else:
                if not b_file_found:
                    b_file_found = True
                    grid_file = argument

        return grid_file



import time, queue
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):


        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################
        specificConfig = self.config.get('SpecificConfig')

        if not specificConfig:

            recordDone = Record(None, message='Trigger: {:04d}'.format(self.config['Identifier']))
            self.return_queue.put(recordDone)
            return

        baseAPI     = None
        pubMethod   = None
        prvMethod   = None
        orderType   = None
        side        = None
        pubSymbol   = None
        prvSymbol   = None
        symbols     = None
        timeframe   = None
        limit       = None     
        amount      = None
        price       = None
        address     = None
        params      = None

        for attrs in specificConfig:
            if attrs['Name'] == 'BaseApi':
                baseAPI = attrs['Data']
            elif attrs['Name'] == 'Public Methods':
                pubMethod = attrs['Data']
            elif attrs['Name'] == 'Private Methods':
                prvMethod = attrs['Data']
            elif attrs['Name'] == 'Order Types':
                orderType = attrs['Data']
            elif attrs['Name'] == 'Side':
                side = attrs['Data']
            elif attrs['Name'] == 'SymbolPublic':
                pubSymbol = attrs['Data']
            elif attrs['Name'] == 'SymbolPrivate':
                prvSymbol = attrs['Data']
            elif attrs['Name'] == 'Timeframe':
                timeframe = attrs['Data']
            elif attrs['Name'] == 'LimitData':
                limit = attrs['Data']
            elif attrs['Name'] == 'Tickers':
                symbols = attrs['Data']
            elif attrs['Name'] == 'Amount':
                amount = attrs['Data']
            elif attrs['Name'] == 'Price':
                price = attrs['Data']
            elif attrs['Name'] == 'Address':
               address = attrs['Data']
            elif attrs['Name'] == 'Parameter':
                params = attrs['Data']
        
        #########################################
        #                                       #
        #    The execution exits immediately    #
        #    after providing output data        #
        #                                       #
        #########################################

        if baseAPI == 'Public':
            methodName  = pubMethod
            symbol      = pubSymbol
        else:
            methodName  = prvMethod
            symbol      = prvSymbol


        if methodName == 'create order' and orderType == 'Market':

            methodName = methodName.replace(" ", "_")
            
            apiCall = {
                'method' : methodName,
                'kwargs' : {
                    'symbol'    : symbol,
                    'type'      : orderType,
                    'side'      : side,
                    'amount'    : amount
                }
            }

        elif methodName == 'create order' and orderType == 'Limit':

            methodName = methodName.replace(" ", "_")

            apiCall = {
                'method' : methodName,
                'kwargs' : {
                    'symbol'    : symbol,
                    'type'      : orderType,
                    'side'      : side,
                    'amount'    : amount,
                    'price'     : price
                }
            }

        elif methodName == 'create order' :

            methodName = methodName.replace(" ", "_")

            apiCall = {
                'method' : methodName,
                'params' : params,
                'kwargs' : {
                    'symbol'    : symbol,
                    'type'      : orderType,
                    'side'      : side,
                    'amount'    : amount,
                    'price'     : price
                }
            }

        elif methodName == 'fetch orders' or        \
             methodName == 'fetch open orders' or   \
             methodName == 'fetch closed orders' or \
             methodName == 'fetch my trades' or     \
             methodName == 'fetch trades' or        \
             methodName == 'fetch order book' or    \
             methodName == 'fetch ticker':
            

            methodName = methodName.replace(" ", "_")

            apiCall = {
                'method' : methodName,
                'kwargs' : {
                    'symbol'    : symbol
                }
            }

        elif methodName == 'fetch tickers':

            
            methodName = methodName.replace(" ", "_")
            apiCall = {
                'method' : methodName,
                'kwargs' : {
                    'symbols'    : symbols
                }
            }

        elif methodName == 'withdraw':

            apiCall = {
                'method' : methodName,
                'kwargs' : {
                    'code'      : symbol,
                    'amount'    : amount,
                    'address'   : address
                }
            }

        elif methodName == 'fetch OHLCV':

            methodName = methodName.replace(" ", "_").lower()

            apiCall = {
                'method' : methodName,
                'kwargs' : {
                    'symbol'    : symbol,
                    'timeframe' : timeframe,
                    'limit'     : int(limit)
                }
            }

        else:

            methodName = methodName.replace(" ", "_")

            apiCall = {
                'method' : methodName
            }

        recordDone = Record(apiCall)     
        self.return_queue.put(recordDone)


import time, queue, sqlite3
try:
    from element_types import Record, Function, ProcCMD, GuiCMD, PythonicError
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):

        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################

        specificConfig = self.config.get('SpecificConfig')

        if not specificConfig:

            recordDone = Record(None, message='Config missing')
            self.return_queue.put(recordDone)
            return

        filename = None
        output = None

        for attrs in specificConfig:
            if attrs['Name'] == 'Filename':
                filename = attrs['Data']

        if self.inputData is None:
            recordDone = Record(None, message='No input provided')
            self.return_queue.put(recordDone)
            return

        con = sqlite3.connect(filename)

        if not con:
            raise Exception('Can not connect to database')

        cur = con.cursor()

        try:
            cur.execute(self.inputData)
        except Exception as e:
            recordDone = Record(PythonicError(e), 'Query failed') 
            self.return_queue.put(recordDone)
            con.close()    
            return

        output = cur.fetchall()

        con.commit()
        con.close()
        recordDone = Record(output, 'Query successful')     
        self.return_queue.put(recordDone)


import setuptools

with open('README.whl', 'r') as fh:
    long_description = fh.read()


setuptools.setup(
        version = '1.10',
        author = 'Stephan Avenwedde',
        author_email = 's.avenwedde@gmail.com',
        license = 'GPLv3',
        description='Graphical automation tool',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/hANSIc99/Pythonic',
        packages = ['Pythonic', 'Pythonic.executables'],
        package_dir = { '' : 'src'},
        package_data = { '' :
            [
             'public_html/static/ManualScheduler.png',
             'public_html/static/Scheduler.png',
             'public_html/static/ManualStopProcess.png',
             'public_html/static/StopProcess.png',
             'public_html/static/GenericPipe.png',
             'public_html/static/GenericProcess.png',
             'public_html/static/ProcessPipe.png',
             'public_html/static/Telegram.png',
             'public_html/static/Email.png',
             'public_html/static/CCXT.png',
             'public_html/static/CCXT_Method.png',
             'public_html/static/SQLite.png',
             'public_html/static/python.ico',
             'public_html/static/qtlogo.svg',
             'public_html/static/qtloader.js',
             'public_html/static/*.js',
             'public_html/static/*.wasm',
             'public_html/static/*.data',
             'public_html/templates/*.html',
             'public_html/config/Toolbox/0Basic/*',
             'public_html/config/Toolbox/1IO/2SQLite.json',
             'public_html/config/Toolbox/1IO/SQLite.editor',
             'public_html/config/Toolbox/2Connectivity/Telegram.editor',
             'public_html/config/Toolbox/2Connectivity/0Telegram.json',
             'public_html/config/Toolbox/2Connectivity/EMail.editor',
             'public_html/config/Toolbox/2Connectivity/1EMail.json',
             'public_html/config/Toolbox/3Trading/*'
             ]
        },
        entry_points = {
            'console_scripts' : ['Pythonic = Pythonic.script:run']
            },
        python_requires = '>=3.7',
        classifiers = [
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Development Status :: 4 - Beta',
            'Intended Audience :: Manufacturing',
            'Intended Audience :: End Users/Desktop',
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Operating System :: OS Independent',
            'Environment :: Web Environment',
            'Environment :: Console',
            'Natural Language :: English',
            'Topic :: Software Development'
            ],
        )



import time, queue
try:
    from element_types import Record, Function, ProcCMD, GuiCMD
except ImportError:    
    from Pythonic.element_types import Record, Function, ProcCMD, GuiCMD
    
class Element(Function):

    def __init__(self, id, config, inputData, return_queue, cmd_queue):
        super().__init__(id, config, inputData, return_queue, cmd_queue)


    def execute(self):


        #####################################
        #                                   #
        #     REFERENCE IMPLEMENTATION      #
        #                                   #
        #####################################



        if self.inputData is None:
            output = 0
        else:
            output = self.inputData + 1


        
        #########################################
        #                                       #
        #    The execution exits immediately    #
        #    after providing output data        #
        #                                       #
        #########################################

        recordDone = Record(output, 'Sending value of cnt: {}'.format(output))     
        self.return_queue.put(recordDone)